from documents.models import PageType, PageDirection, Page, DocumentType, Document, DocumentLink
from images.models import Image, ImageServer
from images.importer import bulk_zones, bulk_transcriptions
from abc import ABC, abstractmethod
import re
import roman
import logging
import os
import fnmatch
import ijson

REGEX_FOLIO_PAGE = re.compile(r'^0*(\d+)(r|v)?')
REGEX_FOLIO_ROMAN = re.compile(r'^0*([MDCLXVI]+)(r|v)?')
FOLIO_WORDS = {
    'contre-plat': PageType.ContrePlat,
    'contre-plat inférieur': PageType.ContrePlat,
    'contre-plat supérieur': PageType.ContrePlat,
    'contre-garde': PageType.ContreGarde,
    'contre-garde supérieur': PageType.ContreGarde,
    'plat': PageType.Plat,
    'plat inférieur': PageType.Plat,
    'plat supérieur': PageType.Plat,
    'garde': PageType.Garde,
    'garde supérieur': PageType.Garde,
    'copie insérée': PageType.Copy,
}


def import_page(volume, image, register, folio, order):
    '''
    Build a page, inside a volume & linked to register+image
    '''
    assert isinstance(volume, Document) and volume.type == DocumentType.Volume
    assert isinstance(image, Image)

    # Create a single register document holding all the pages
    document, _ = Document.objects.get_or_create(
        type=DocumentType.Register,
        name=register,
    )
    DocumentLink.objects.get_or_create(parent=volume, child=document)

    # Build the page
    page_type, page_nb, page_direction = parse_folio(folio)
    p, _ = Page.objects.get_or_create(image=image, defaults={
        'folio': folio,
        'name': "Page {0} du volume {1}".format(folio, volume.name),
        'page_type': page_type,
        'nb': page_nb,
        'direction': page_direction
    })
    DocumentLink.objects.get_or_create(parent=document, child=p, defaults={'order': order})
    return p


def parse_folio(folio):
    """
    Parse folio to build a DB representation
    Outputs (type, nb, direction)
    """
    if folio is None:
        return (None, None, None)

    def _direction(d):
        if d is None:
            return
        if d.lower() in ('r', 'recto'):
            return PageDirection.Recto
        if d.lower() in ('v', 'verso'):
            return PageDirection.Verso

    # Standard page
    # Removing zeros, because this would match even roman numerals
    page_match = REGEX_FOLIO_PAGE.match(folio.lstrip('0'))
    if page_match is not None:
        groups = page_match.groups()
        page = int(groups[0])
        direction = _direction(groups[1])
        return (PageType.Page, page, direction)

    # Roman page
    roman_match = REGEX_FOLIO_ROMAN.match(folio)
    if roman_match is not None:
        groups = roman_match.groups()
        page = roman.fromRoman(groups[0])
        direction = _direction(groups[1])
        return (PageType.Summary, page, direction)

    # Words
    page_type = FOLIO_WORDS.get(folio.lower())

    return (page_type, None, None)


class ManifestsImporter(ABC):
    """
    Parses JSON manifests and annotation data to import them in the database.
    """

    def __init__(self, imgserv):
        """Initialize a manifest importer
        `imgserv` can be either one ImageServer or a list of ImageServers."""
        if isinstance(imgserv, ImageServer):
            self.imgserv = [imgserv]
        else:
            assert all(isinstance(srv, ImageServer) for srv in imgserv)
            self.imgserv = imgserv
        self.logger = logging.getLogger(__name__)
        # This dictionary associates canvas IDs with images
        # Filled by parse_manifest ; used by parse_annotation_list
        self.canvas_image = {}
        # This dictionary holds parsed data for each image's zones and transcriptions
        # Filled by parse_annotation_list ; used by save_transcriptions
        self.images_transcription_data = {}

    @abstractmethod
    def get_json_files(self):
        """
        Generator yielding a stream (any class that has read() and seek(int) methods)
        for each JSON file to parse
        """
        pass

    def run(self):
        """Start the importing process."""
        for f in self.get_json_files():
            self.parse_file(f)
        self.save_transcriptions()

    def parse_file(self, stream):
        """Parse a JSON file loaded in a stream."""
        parser = ijson.parse(stream)
        for prefix, event, value in parser:
            if (prefix, event) == ('@type', 'string'):
                if value == 'sc:Manifest':
                    stream.seek(0)
                    self.parse_manifest(stream)
                    break
                elif value == 'sc:AnnotationList':
                    stream.seek(0)
                    self.parse_annotation_list(stream)
                    break

    def parse_manifest(self, stream):
        """Parse a IIIF manifest loaded as a stream."""
        # Get this file's volume range ID from the top-most structure
        range_id = next(struct['ranges'][0]
                        for struct in ijson.items(stream, "structures.item")
                        if struct.get('viewingHint') == "top")
        stream.seek(0)

        # Get our volume's structure and abel
        vol_struct = next(struct for struct in ijson.items(stream, "structures.item")
                          if struct.get('@id') == range_id)
        vol_name = vol_struct['label']
        self.logger.debug("Creating volume {}".format(vol_name))

        # Create a volume and a register
        vol, _ = Document.objects.get_or_create(name=vol_name, type=DocumentType.Volume)
        doc, _ = Document.objects.get_or_create(name=vol_name, type=DocumentType.Register)
        DocumentLink.objects.get_or_create(parent=vol, child=doc)
        stream.seek(0)
        self.parse_manifest_canvases(stream, vol)

    def parse_manifest_canvases(self, stream, volume):
        """Parse canvases in a IIIF manifest stream and
        create images and pages with a given volume name."""
        # Enumerating to use index as order number in DocumentLink
        for i, canvas in enumerate(ijson.items(stream, 'sequences.item.canvases.item')):
            # Label contains the folio
            if canvas.get('label') is None:
                self.logger.warning(
                    "Found an image canvas with id {0} for volume {1}, "
                    "but no folio (label) was specified".format(
                        canvas.get('@id'), str(volume)))
                continue
            folio = canvas['label']

            # Get the image resource
            resource = canvas['images'][0].get('resource')
            if resource is None:
                self.logger.warning(
                    "Found an image canvas with id {0} for volume {1}, "
                    "but no image resource was specified".format(
                        canvas.get('@id'), str(volume)))
                continue

            # Go find the service ID to get the image URL
            if resource.get('service') is None:
                self.logger.warning(
                    "Found an image resource with id {}, but no service was specified".format(
                        resource.get('@id')))
                continue
            service_id = resource['service'].get('@id')

            # Find the right server
            image_server = next(srv for srv in self.imgserv if service_id.startswith(srv.url))
            # Strip server URL to get image path
            path = service_id[len(image_server.url):].lstrip('/')

            # Create image and page
            self.logger.debug("Creating image {}".format(path))
            image = image_server.find_image(path)
            self.logger.debug("Creating page {}".format(folio))
            import_page(volume, image, volume.name, folio, i)

            # Fill canvas_image dictionary for usage by annotation list parsing
            self.canvas_image[canvas.get('@id')] = image

    def parse_annotation_list(self, stream):
        """Parse a IIIF annotation list loaded as a stream.
        Zones and transcriptions are not immediately saved; use save_transcriptions()."""
        for annotation in ijson.items(stream, 'resources.item'):
            # Check annotation contains text
            if annotation['resource'].get('chars') is None or \
                    annotation['resource']['@type'] != 'cnt:ContentAsText' or \
                    annotation['resource']['format'] != 'text/plain':
                self.logger.warning("Incompatible annotation with ID {}".format(
                    annotation.get('@id')))
                continue
            text = annotation['resource']['chars']

            # Get ID and zone information from URI
            canvas_id, data = annotation['on'].split('#', 1)

            # Find the associated image
            if canvas_id not in self.canvas_image.keys():
                self.logger.warning(
                    "Could not find an associated image for annotation with ID {}".format(
                        annotation.get('@id')))
                continue
            image = self.canvas_image[canvas_id]
            # Add to dictionary
            if self.images_transcription_data.get(image) is None:
                self.images_transcription_data[image] = []

            # Parse the data string (format is xywh=0,100,200,300) and cast to int
            # rpartition("=")[2] gives the part after the '=' and is faster than a split
            x, y, w, h = [int(v) for v in data.rpartition('=')[2].split(',')]

            # Fill data dictionary
            self.images_transcription_data[image].append({
                'x': x, 'y': y, 'width': w, 'height': h, 'text': text, 'line': None, 'score': 1
            })

    def save_transcriptions(self):
        """To optimize transcription parsing, saving and indexing is done in bulk."""
        total_zones, total_transcriptions, total_indexes = 0, 0, 0

        for image, data in self.images_transcription_data.items():
            new_zones = bulk_zones(image, data)
            total_zones += len(new_zones)
            self.logger.debug("Created {0} zones for image {1}".format(len(new_zones), image.path))

            new_transcriptions = bulk_transcriptions(image, data)
            total_transcriptions += len(new_transcriptions)
            self.logger.debug("Created {0} transcriptions for image {1}".format(len(new_transcriptions), image.path))

            index_count = image.index()
            total_indexes += index_count
            self.logger.debug("Created {0} indexes for image {1}".format(index_count, image.path))

        self.logger.info("Saved {0} zones, {1} transcriptions and {2} indexes".format(
            total_zones, total_transcriptions, total_indexes))


class LocalManifestsImporter(ManifestsImporter):
    """Allows importing of local JSON files."""

    def __init__(self, imgserv, folder):
        super().__init__(imgserv)
        assert os.path.isdir(folder), 'Invalid folder path {}'.format(folder)
        self.folder = folder

    def get_json_files(self):
        for root, _, filenames in os.walk(self.folder):
            for filename in fnmatch.filter(filenames, "*.json"):
                self.logger.debug("Opening {}".format(filename))
                yield open(os.path.join(root, filename), 'rb')
