from django.db import models
from enumfields import EnumField, Enum
from arkindex.models import IndexableModel
from documents.managers import DocumentManager


class DocumentType(Enum):
    Volume = 'volume'
    Register = 'register'
    Act = 'act'
    Page = 'page'


class DocumentLink(models.Model):
    """
    Links documents together through a parent-child relation
    """
    parent = models.ForeignKey('documents.Document', related_name='link_parent')
    child = models.ForeignKey('documents.Document', related_name='link_child')
    order = models.PositiveIntegerField(default=0)

    class Meta:
        unique_together = (
            ('parent', 'order'),
            ('parent', 'child'),
        )

    def __str__(self):
        return "'{}' -> '{}'".format(self.parent, self.child)


class Document(IndexableModel):
    """
    aka Unité Codicologique
    """
    parents = models.ManyToManyField('self', through=DocumentLink, symmetrical=False)
    type = EnumField(DocumentType, max_length=50)
    name = models.CharField(max_length=250)
    objects = DocumentManager()

    def __str__(self):
        return '{} : {}'.format(self.type, self.name)


class PageType(Enum):
    Page = 'page'
    Summary = 'summary'
    Garde = 'garde'
    ContreGarde = 'contregarde'
    ContrePlat = 'contreplat'
    Plat = 'plat'
    Copy = 'copy'


class PageDirection(Enum):
    Recto = 'recto'
    Verso = 'verso'


class Page(Document):
    """
    with folio numbering
    """
    folio = models.CharField(max_length=250)

    # Parsed folio
    page_type = EnumField(PageType, max_length=50, null=True, blank=True)
    nb = models.PositiveIntegerField(null=True, blank=True)
    direction = EnumField(PageDirection, max_length=10, null=True, blank=True)

    # The current image being used
    image = models.OneToOneField('images.Image', related_name='page')

    def save(self, *args, **kwargs):
        # TODO: move this in Document through introspection
        self.type = DocumentType.Page
        super().save(*args, **kwargs)

    def __str__(self):
        out = [self.type.value, ]
        if self.nb:
            out.append('n°{}'.format(self.nb))
        if self.direction:
            out.append(self.direction.value)
        return ' '.join(out)


class Section(Document):
    """
    A logical subdivision that can span on multiple images via zones
    """
    images = models.ManyToManyField('images.Image', through='images.Zone')

    def save(self, *args, **kwargs):
        # TODO: move this in Document through introspection
        self.type = DocumentType.Act
        super().save(*args, **kwargs)


class Transcription(IndexableModel):
    """
    A transcription on:
     * a zone on an image
    """
    INDEX_TYPE = 'transcription'
    zone = models.ForeignKey('images.Zone', related_name='transcriptions')

    # Data
    line = models.PositiveIntegerField(null=True, blank=True)
    text = models.TextField(null=True, blank=True)
    score = models.FloatField(null=True, blank=True)

    class Meta:
        unique_together = (
            ('zone', 'text'),
        )

    def __str__(self):
        return 'Transcription: {}'.format(self.text[:20])

    def build_search_index(self):
        """
        Structure indexed into Elastic Search
        """
        return {
            'image': self.zone.image_id.hex,
            'zone': self.zone_id.hex,
            'score': self.score,
            'line': self.line,
            'text': self.text,
        }
