from documents.models import Transcription, Page
from django.conf import settings
from elasticsearch.helpers import scan as es_scan
from elasticsearch import Elasticsearch
from images import ES_INDEX
from images.models import Zone
import itertools


def search_transcriptions(query, min_score=0.5):
    """
    Search transcriptions indexed into ElasticSearch
    Only returns Transcriptions above specified score
    No Postgresql hits
    """
    query = {
        'bool': {
            'must': [
                {
                    'match': {
                        'text': query,
                    }
                },
                {
                    'range': {
                        'score': {
                            'gte': min_score,
                        }
                    }
                },
            ]
        }
    }
    elastic = Elasticsearch(settings.ELASTIC_SEARCH_HOSTS)
    results = es_scan(
        client=elastic,
        index=ES_INDEX,
        query={'query': query},
        doc_type=Transcription.INDEX_TYPE,
    )

    return results


def search_pages(query):
    '''
    Search pages with a transcription
    List pages ordered by document
    '''
    assert isinstance(query, str)
    assert query != ''

    # Query indexed transcriptions
    results = search_transcriptions(query)

    # Group by images, as an in-memory dict
    # dict keys are image ids
    def grouper(r): return r['_source']['image']
    results = sorted(results, key=grouper)
    indexes = {
        image_id: list(indexes)
        for image_id, indexes in itertools.groupby(results, grouper)
    }

    # Load Pages from results
    pages = Page.objects.filter(image_id__in=indexes.keys())

    # Load needed transcriptions
    transcription_ids = [r['_id'] for r in results]
    transcriptions = {
        t.id.hex: t
        for t in Transcription.objects.filter(id__in=transcription_ids)
    }

    # Load needed zones
    zone_ids = [r['_source']['zone'] for r in results]
    zones = {
        z.id.hex: z
        for z in Zone.objects.filter(id__in=zone_ids).distinct()
    }

    def grouper(r): return r['_source']['zone']
    for page in pages:

        # Group image results by zone
        results = sorted(indexes[page.image_id.hex], key=grouper)
        page.indexes = [
            {
                'zone': zones[zone_id],
                'transcriptions': [
                    transcriptions[r['_id']]
                    for r in res
                ],
            }
            for zone_id, res in itertools.groupby(results, grouper)
        ]

    return pages
