from django.db import models
from django.conf import settings
from elasticsearch import Elasticsearch
from elasticsearch.helpers import bulk as es_bulk
from documents.models import Transcription, Section
from arkindex.models import IndexableModel
from images import ES_INDEX
import requests
import urllib.parse


class ImageServer(models.Model):
    """
    A IIIF image server
    """
    name = models.CharField(max_length=250)
    url = models.URLField()

    created = models.DateTimeField(auto_now_add=True)
    updated = models.DateTimeField(auto_now=True)

    def __str__(self):
        return self.name

    def find_image(self, path):
        """
        Lookup an image on server
        This is the preferred way to construct an image
        """
        # Try direct access to path
        try:
            return self.images.get(path=path)
        except Image.DoesNotExist:
            pass

        # Try the url encoded path
        try:
            return self.images.get(
                path=urllib.parse.quote_plus(path),
            )
        except Image.DoesNotExist:
            pass

        # Check the source
        if not path.endswith('/'):
            path += '/'
        info_url = urllib.parse.urljoin(self.build_url(path), 'info.json')
        resp = requests.get(info_url, allow_redirects=True)
        resp.raise_for_status()

        # Use Image id from IIIF server
        image_id = resp.json().get('@id')
        assert image_id is not None, \
            'Missing image id in server response'
        assert image_id.startswith(self.url), \
            'Image id does not start with server url ({} vs. {})'.format(
                image_id, self.url)
        image_path = image_id[len(self.url) + 1:]

        # Create new image
        return self.images.create(
            path=image_path,
        )

    def build_url(self, path):
        """
        Build a full url on this server
        """
        base_url = self.url
        if not base_url.endswith('/'):
            base_url += '/'
        return urllib.parse.urljoin(base_url, urllib.parse.quote(path.lstrip('/')))


class Image(IndexableModel):
    """
    A document image
    """
    server = models.ForeignKey(ImageServer, related_name='images')
    path = models.URLField()

    class Meta:
        unique_together = (
            ('server', 'path'),
        )

    @property
    def url(self):
        return self.server.build_url(self.path)

    def __str__(self):
        return '{} - {}'.format(self.id, self.url)

    def index(self):
        """
        Update the transcriptions payload into ElasticSearch
        """
        assert self.zones.exists(), \
            'No zones to index'

        # Load directly all transcriptions
        transcriptions = Transcription.objects.filter(
            zone__image_id=self.id,
        ).prefetch_related('zone')

        # Build raw ElasticSearch insert
        actions = [{
            '_index': ES_INDEX,
            '_type': Transcription.INDEX_TYPE,
            '_source': t.build_search_index(),
            '_id': t.id.hex,
        } for t in transcriptions.all()]

        # Run actions in bulk
        elastic = Elasticsearch(settings.ELASTIC_SEARCH_HOSTS)
        nb_insert, _ = es_bulk(elastic, actions)

        return nb_insert


class Zone(IndexableModel):
    """
    A zone on an image
    """
    image = models.ForeignKey(Image, related_name='zones')
    section = models.ForeignKey(Section, related_name='zones',
                                null=True, blank=True)

    # Box
    # (x,y) is top left corner -> raster
    x = models.PositiveIntegerField()
    y = models.PositiveIntegerField()
    width = models.PositiveIntegerField()
    height = models.PositiveIntegerField()

    def __str__(self):
        return 'Zone ({},{}) w={} h={}'.format(self.x, self.y, self.width, self.height)

    class Meta:
        unique_together = (
            ('image', 'x', 'y', 'width', 'height'),
        )

    @property
    def url(self):
        '''
        Build the IIIF crop url
        '''
        return urllib.parse.urljoin(
            self.image.url + '/',
            '{},{},{},{}/full/0/default.jpg'.format(
                self.x, self.y, self.width, self.height),
        )
